@@ -84,6 +84,7 @@ class CoordinateData(Protocol):
     geometry: Geometry
     axes: tuple[Name, ...]
     coordinates: FieldMap
+    fields: FieldMap
 
 
 class ValidatorMixin(GeometricData, ABC):
@@ -139,8 +140,8 @@ class ValidatorMixin(GeometricData, ABC):
     def _validate_geometry(self) -> None:
         known_axes: dict[Geometry, tuple[Name, Name, Name]] = {
             Geometry.CARTESIAN: ("x", "y", "z"),
-            Geometry.POLAR: ("radius", "z", "azimuth"),
-            Geometry.CYLINDRICAL: ("radius", "azimuth", "z"),
+            Geometry.POLAR: ("radius", "azimuth", "z"),
+            Geometry.CYLINDRICAL: ("radius", "z", "azimuth"),
             Geometry.SPHERICAL: ("radius", "colatitude", "azimuth"),
             Geometry.EQUATORIAL: ("radius", "azimuth", "latitude"),
         }
@@ -151,11 +152,13 @@ class ValidatorMixin(GeometricData, ABC):
                 f"Unknown geometry {self.geometry.name.lower()!r}"
             )  # pragma: no cover
 
-        axes = known_axes[self.geometry]
-        for expected, actual in zip(axes, self.axes):
+        axes = known_axes[self.geometry][: len(self.axes)]
+        for i, (expected, actual) in enumerate(zip(axes, self.axes)):
             if actual != expected:
                 raise ValueError(
-                    f"Got invalid axis {actual!r} with geometry {self.geometry.name.lower()!r}"
+                    f"Got invalid axis name {actual!r} on position {i}, "
+                    f"with geometry {self.geometry.name.lower()!r}\n"
+                    f"Expected axes ordered as {axes}"
                 )
 
 
@@ -170,30 +173,48 @@ _AXES_LIMITS: dict[Name, tuple[float, float]] = {
 
 
 class CoordinateValidatorMixin(ValidatorMixin, CoordinateData, ABC):
+    def __init__(self) -> None:
+        super().__init__()
+        dts = {
+            name: arr.dtype
+            for name, arr in chain(
+                self.coordinates.items(),
+                self.fields.items(),
+            )
+        }
+        unique_dts = sorted({dtype for dtype in dts.values()})
+        if len(unique_dts) > 1:
+            raise TypeError(f"Grid received mixed data types ({unique_dts}):\n{dts}")
+        else:
+            self.dtype = unique_dts.pop()
+
     def _validate_coordinates(self) -> None:
         for axis in self.axes:
             coord = self.coordinates[axis]
             if len(coord) == 0:
                 continue
-            xmin, xmax = _AXES_LIMITS[axis]
-            if (cmin := np.min(coord)) < xmin:
+            coord_dtype = self._get_safe_datatype(coord)
+            dt = coord_dtype.type
+            xmin, xmax = (dt(_) for _ in _AXES_LIMITS[axis])
+            if (cmin := dt(np.min(coord))) < xmin:
                 raise ValueError(
                     f"Invalid coordinate data for axis {axis!r} {cmin} "
                     f"(minimal allowed value is {xmin})"
                 )
-            if (cmax := np.max(coord)) > xmax:
+            if (cmax := dt(np.max(coord))) > xmax:
                 raise ValueError(
                     f"Invalid coordinate data for axis {axis!r} {cmax} "
                     f"(maximal allowed value is {xmax})"
                 )
 
-            self.coordinates[axis] = coord.astype(self._get_safe_datatype(), copy=False)
+            self.coordinates[axis] = coord.astype(coord_dtype, copy=False)
 
-    def _get_safe_datatype(self) -> np.dtype:
+    def _get_safe_datatype(self, reference: np.ndarray | None = None) -> np.dtype:
         # int32 and int64 are fragile because they cannot represent "+/-inf",
         # which we use as default box boundaries, e.g. for cartesian datasets
-
-        dt = self.coordinates[self.axes[0]].dtype
+        if reference is None:
+            reference = self.coordinates[self.axes[0]]
+        dt = reference.dtype
         dt_str = str(dt)
         if dt_str.startswith("int"):
             return np.dtype(f"float{dt_str[3:]}")
@@ -364,28 +385,11 @@ class Dataset(ValidatorMixin):
                 if x > domain_right:
                     raise ValueError(f"Got particle at {ax}={x} > {domain_right=}")
 
-        dts_grid = {
-            name: arr.dtype
-            for name, arr in chain(
-                self.grid.coordinates.items(),
-                self.grid.fields.items(),
-            )
-        }
-        dts_particles = {
-            name: arr.dtype
-            for name, arr in chain(
-                self.particles.coordinates.items(),
-                self.particles.fields.items(),
-            )
-        }
-        unique_dts = sorted(
-            {dtype for dtype in chain(dts_grid.values(), dts_particles.values())}
-        )
-        if len(unique_dts) > 1:
+        if self.particles.dtype != self.grid.dtype:
             raise TypeError(
-                f"Got mixed data types ({unique_dts})\n"
-                f"- from grid data: {dts_grid}\n"
-                f"- from particles: {dts_particles}\n"
+                f"Got mixed data types:\n"
+                f"- from grid data: {self.grid.dtype}\n"
+                f"- from particles: {self.particles.dtype}\n"
             )
 
     def _get_padded_cell_edges(self) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
@@ -436,6 +440,13 @@ class Dataset(ValidatorMixin):
             dx=self.grid._dx,
             out=self._hci,
         )
+        for idim in range(self._hci.shape[1]):
+            # There are at leas two edge cases where we need clipping to correct raw indices:
+            # - particles that live exactly on the domain right edge will be indexed out of bound
+            # - single precision positions can lead to overshoots by simple effect of floating point arithmetics
+            #   (double precision isn't safe either, it's just safer (by a lot))
+            self._hci[:, idim].clip(1, self.grid.shape[idim], out=self._hci[:, idim])
+
         tstop = monotonic_ns()
         if verbose:
             print(
