@@ -25,8 +25,8 @@ def test_cell_volumes_curvilinear():
         geometry="cylindrical",
         grid={
             "cell_edges": {
-                "radius": np.arange(11),
-                "azimuth": np.arange(0, 1.1, 0.1),
+                "radius": np.arange(11, dtype="int64"),
+                "z": np.arange(0, 1.1, 0.1),
             }
         },
     )
@@ -34,4 +34,4 @@ def test_cell_volumes_curvilinear():
         NotImplementedError,
         match=r"cell_volumes property is not implemented for cylindrical geometry",
     ):
-        ds.grid.cell_volumes
+        ds.grid.cell_volumes  # noqa: B018
